import os
from collections import defaultdict
from os.path import join as pjoin
from pathlib import Path

import pysam
import pytest

from ugbio_featuremap.featuremap_consensus_utils import pileup_featuremap, pileup_featuremap_on_an_interval_list


@pytest.fixture
def resources_dir():
    return Path(__file__).parent / "resources"


@pytest.mark.parametrize(
    "genomic_interval, interval_list, expected_num_variants",
    [
        ("chr19:1-400000", None, 66),
        (None, None, 477),
        ("chr2:1-1000", None, 0),
        (None, pjoin(Path(__file__).parent / "resources", "scattered.interval_list"), 477),
    ],
)
def test_pileup_featuremap(
        tmpdir, resources_dir,
        genomic_interval,
        interval_list,
        expected_num_variants,
):
    # create input featuremap vcf file
    input_featuremap_vcf = pjoin(resources_dir,
                                 "HG001_HG002_tumor_tumor_in_normal_31977.gtr.hcr.chr19_1-1000000.vcf.gz")

    # call the function with different arguments
    pileup_featuremap_vcf = pjoin(tmpdir, "pileup_featuremap.vcf.gz")
    if interval_list is None:
        pileup_featuremap(
            featuremap=input_featuremap_vcf,
            output_vcf=pileup_featuremap_vcf,
            genomic_interval=genomic_interval,
        )
    else:
        pileup_featuremap_on_an_interval_list(
            featuremap=input_featuremap_vcf,
            output_vcf=pileup_featuremap_vcf,
            interval_list=interval_list,
        )

    # check that the output file exists and has the expected content
    assert os.path.isfile(pileup_featuremap_vcf)
    # count the number of variants (excluding the header)
    cons_dict = defaultdict(dict)
    for rec in pysam.VariantFile(pileup_featuremap_vcf):
        rec_id = (rec.chrom, rec.pos, rec.ref, rec.alts[0])
        if rec_id not in cons_dict:
            cons_dict[rec_id]["count"] = 0
        cons_dict[rec_id]["count"] += 1
    # check the number of variants
    num_variants = len(cons_dict)
    assert num_variants == expected_num_variants
    # assert a single entry per variant
    assert all(cons["count"] == 1 for cons in cons_dict.values())



#!/env/python
# Copyright 2023 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Create a HOM SNV featuremap from a featuremap
# CHANGELOG in reverse chronological order
import argparse
import sys

from simppl.simple_pipeline import SimplePipeline

from ugbio_featuremap.featuremap_utils import create_hom_snv_featuremap


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="create_hom_snv_featuremap.py", description=run.__doc__)
    parser.add_argument(
        "--featuremap",
        type=str,
        required=True,
        help="""Input featuremap""",
    )
    parser.add_argument(
        "--sorter_stats_json",
        type=str,
        default=None,
        help="Path to Sorter statistics JSON file. If None, the minimum coverage "
        "will be set to requested_min_coverage even if the median coverage is lower, might yield an empty output.",
    )
    parser.add_argument(
        "--hom_snv_featuremap",
        type=str,
        default=None,
        help="Output featuremap with HOM SNVs reads to be used as True Positives. If None (default), "
        "the hom_snv_featuremap will be the same as the input featuremap with a '.hom_snv.vcf.gz' suffix.",
    )
    parser.add_argument(
        "--requested_min_coverage",
        type=int,
        default=20,
        help="Minimum coverage requested for locus to be propagated to the output. If the median coverage "
        "is lower than this value, the median coverage will be used as the minimum coverage instead.",
    )
    parser.add_argument(
        "--min_af",
        type=float,
        default=0.7,
        help="Minimum allele fraction in the featuremap to be considered a HOM SNV"
        "The default is chosen as 0.7 and not higher because some SNVs are pre-filtered from the FeatureMap due to "
        "MAPQ<60 or due to adjacent hmers.",
    )

    return parser


def run(argv):
    """Create a HOM SNV featuremap from a featuremap"""
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)

    create_hom_snv_featuremap(
        featuremap=args.featuremap,
        sorter_stats_json=args.sorter_stats_json,
        hom_snv_featuremap=args.hom_snv_featuremap,
        sp=sp,
        requested_min_coverage=args.requested_min_coverage,
        min_af=args.min_af,
    )

def main():
    run(sys.argv)


if __name__ == "__main__":
    main()


from collections import defaultdict
from os.path import join as pjoin
from pathlib import Path

import pysam
import pytest

from ugbio_featuremap import create_hom_snv_featuremap

@pytest.fixture
def resources_dir():
    return Path(__file__).parent / "resources"


def test_create_hom_snv_featuremap(tmpdir, resources_dir):
    # this vcf spans 5 HOM SNV loci, out of which 3 meet the criteria, 1 is low coverage (10) and 1 is low AF (9/30)
    hom_snv_featuremap = pjoin(tmpdir, "Pa_46_333_LuNgs_08.chr1_1000291_1002833.hom_snv.vcf.gz")
    create_hom_snv_featuremap.run(
        [
            "create_hom_snv_featuremap.py",
            "--featuremap",
            pjoin(resources_dir, "Pa_46_333_LuNgs_08.chr1_1000291_1002833.vcf.gz"),
            "--sorter_stats_json",
            pjoin(resources_dir, "Pa_46_333_LuNgs_08.json"),
            "--hom_snv_featuremap",
            hom_snv_featuremap,
            "--requested_min_coverage",
            "20",
            "--min_af",
            "0.7",
        ]
    )

    # making sure the output is as expected, with the 3 correct loci
    loci_counter = defaultdict(int)
    with pysam.VariantFile(hom_snv_featuremap) as fvcf:
        for record in fvcf.fetch():
            loci_counter[(record.chrom, record.pos)] += 1
    assert loci_counter[("chr1", 1002308)] == 59
    assert loci_counter[("chr1", 1002736)] == 56
    assert loci_counter[("chr1", 1002833)] == 54
    assert len(loci_counter) == 3


from os.path import join as pjoin
from pathlib import Path

import pysam
import pytest

from ugbio_featuremap.featuremap_utils import FeatureMapFields, annotate_featuremap
from ugbio_ppmseq.ppmSeq_utils import HistogramColumnNames

@pytest.fixture
def resources_dir():
    return Path(__file__).parent / "resources"


def test_annotate_featuremap(tmpdir, resources_dir):
    input_featuremap = pjoin(resources_dir, "tp_featuremap_chr20.vcf.gz")
    output_featuremap = pjoin(tmpdir, "tp_featuremap_chr20.annotated.vcf.gz")
    ref_fasta = pjoin(resources_dir, "sample.fasta")
    annotate_featuremap(
        input_featuremap,
        output_featuremap,
        ref_fasta=ref_fasta,
        ppmSeq_adapter_version="legacy_v5",
        flow_order="TGCA",
        motif_length_to_annotate=3,
        max_hmer_length=20,
    )
    out = pysam.VariantFile(output_featuremap)

    for info_field in [
        "X_CIGAR",
        "X_EDIST",
        "X_FC1",
        "X_FC2",
        "X_READ_COUNT",
        "X_FILTERED_COUNT",
        "X_FLAGS",
        "X_LENGTH",
        "X_MAPQ",
        "X_INDEX",
        "X_RN",
        "X_SCORE",
        "rq",
        FeatureMapFields.IS_FORWARD.value,
        FeatureMapFields.IS_DUPLICATE.value,
        FeatureMapFields.MAX_SOFTCLIP_LENGTH.value,
        HistogramColumnNames.STRAND_RATIO_CATEGORY_START.value,
        HistogramColumnNames.STRAND_RATIO_CATEGORY_END.value,
        HistogramColumnNames.STRAND_RATIO_START.value,
        HistogramColumnNames.STRAND_RATIO_END.value,
        "trinuc_context_with_alt",
        "hmer_context_ref",
        "hmer_context_alt",
        "is_cycle_skip",
        "prev_1",
        "prev_2",
        "prev_3",
        "next_1",
        "next_2",
        "next_3",
    ]:
        assert str(info_field) in out.header.info



import os
from os.path import join as pjoin
from pathlib import Path
import pandas as pd
import pysam
import pytest
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_featuremap.featuremap_utils import filter_featuremap_with_bcftools_view, featuremap_to_dataframe, \
    FeatureMapFields, FeaturemapAnnotator, RefContextVcfAnnotator
from ugbio_core.variant_annotation import VcfAnnotator
import ugbio_core.variant_annotation as variant_annotation
import numpy as np


@pytest.fixture
def resources_dir():
    return Path(__file__).parent / "resources"


@pytest.mark.parametrize(
    "min_coverage, max_coverage, bcftools_include_filter, regions_string, expected_num_variants",
    [
        (None, None, None, "chr20\t85258\t85260\n", 79),
        (10, None, None, "chr20\t85258\t85260\n", 79),
        (None, 100, None, "chr20\t85258\t85260\n", 79),
        (None, None, "(X_SCORE >= 10)", "chr20\t85258\t85260\n", 79),
        (None, 100, "(X_SCORE >= 4) && (X_EDIST <= 5)", "chr20\t85258\t85260\n", 78),
        (None, 100, "(X_SCORE >= 4)", "chr20\t85258\t96322\n", 214),
        (84, 100, "(X_SCORE >= 4)", "chr20\t85258\t96322\n", 79),
    ],
)
def test_filter_featuremap_with_bcftools_view_with_params(
        tmpdir,
        min_coverage,
        max_coverage,
        bcftools_include_filter,
        regions_string,
        expected_num_variants,
        resources_dir
):
    # create input featuremap vcf file
    input_featuremap_vcf = pjoin(resources_dir, "333_LuNgs_08.annotated_featuremap.vcf.gz")

    # create regions file
    regions_file = pjoin(tmpdir, "regions.bed")
    with open(regions_file, "w") as f:
        f.write(regions_string)

    # call the function with different arguments
    intersect_featuremap_vcf = pjoin(tmpdir, "intersect_featuremap.vcf.gz")
    filter_featuremap_with_bcftools_view(
        input_featuremap_vcf=input_featuremap_vcf,
        intersect_featuremap_vcf=intersect_featuremap_vcf,
        min_coverage=min_coverage,
        max_coverage=max_coverage,
        regions_file=regions_file,
        bcftools_include_filter=bcftools_include_filter,
    )

    # check that the output file exists and has the expected content
    assert os.path.isfile(intersect_featuremap_vcf)
    # count the number of variants (excluding the header)
    num_variants = 0
    for _ in pysam.VariantFile(intersect_featuremap_vcf):
        num_variants += 1
    assert num_variants == expected_num_variants


def test_featuremap_to_dataframe(tmpdir, resources_dir):
    input_featuremap = pjoin(resources_dir, "Pa_46.bsDNA.chr20_sample.vcf.gz")
    expected_featuremap_dataframe = pjoin(resources_dir, "Pa_46.bsDNA.chr20_sample.parquet")
    tmp_out_path = pjoin(tmpdir, "tmp_out.parquet")
    featuremap_dataframe = featuremap_to_dataframe(featuremap_vcf=input_featuremap, output_file=tmp_out_path)
    featuremap_dataframe_expected = pd.read_parquet(expected_featuremap_dataframe)
    _assert_read_signature(
        featuremap_dataframe,
        featuremap_dataframe_expected,
        expected_columns=[
            "chrom",
            "pos",
            "ref",
            "alt",
            "qual",
            "filter",
            "MLEAC",
            "MLEAF",
            "X_CIGAR",
            "X_EDIST",
            "X_FC1",
            "X_FC2",
            "X_FILTERED_COUNT",
            "X_FLAGS",
            "X_INDEX",
            "X_LENGTH",
            "X_MAPQ",
            "X_READ_COUNT",
            "X_RN",
            "X_SCORE",
            "X_SMQ_LEFT",
            "X_SMQ_LEFT_MEAN",
            "X_SMQ_RIGHT",
            "X_SMQ_RIGHT_MEAN",
            "a3",
            "ae",
            "as",
            "rq",
            "s2",
            "s3",
            "te",
            "tm",
            "ts",
        ],
        possibly_null_columns=["tm", "filter", "qual", "MLEAC", "MLEAF"],
    )


def _assert_read_signature(signature, expected_signature, expected_columns=None, possibly_null_columns=None):
    expected_columns = expected_columns or [
        "ref",
        "alt",
        "id",
        "qual",
        "af",
    ]
    possibly_null_columns = possibly_null_columns or [
        "id",
        "qual",
    ]
    for c in expected_columns:
        assert c in signature.columns
        if c not in possibly_null_columns:
            assert not signature[c].isnull().all()
            assert (signature[c] == expected_signature[c]).all() or np.allclose(signature[c], expected_signature[c])


def test_featuremap_annotator(tmpdir, resources_dir):
    input_featuremap = pjoin(resources_dir, "Pa_46.bsDNA.chr20_sample.vcf.gz")
    tmpfile = f"{tmpdir}/output_featuremap.vcf.gz"
    VcfAnnotator.process_vcf(
        input_path=input_featuremap,
        output_path=tmpfile,
        annotators=[FeaturemapAnnotator()],
    )
    output_variants = pysam.VariantFile(tmpfile)
    forward_events = 0
    reverse_events = 0
    total_max_softclip_bases = 0
    for v in output_variants:
        assert FeatureMapFields.MAX_SOFTCLIP_LENGTH.value in v.info
        total_max_softclip_bases += int(v.info[FeatureMapFields.MAX_SOFTCLIP_LENGTH.value])
        if FeatureMapFields.MAX_SOFTCLIP_LENGTH.IS_FORWARD.value in v.info:
            forward_events += 1
        else:
            reverse_events += 1
    assert (
            FeatureMapFields.IS_DUPLICATE.value in output_variants.header.info
    ), f"{FeatureMapFields.IS_DUPLICATE.value} is not in info header {output_variants.header.info}"
    assert forward_events == 9
    assert reverse_events == 22
    assert total_max_softclip_bases == 81


def test_ref_context_vcf_annotator_somatic_dv(tmpdir, resources_dir):
    # data files
    sample_vcf = pjoin(resources_dir, "Pa_46.FFPE.chr20_sample.vcf.gz")
    motif_length = 3
    ref_contetxt_variant_annotator = RefContextVcfAnnotator(
        ref_fasta=pjoin(resources_dir, "sample.fasta"),
        flow_order=DEFAULT_FLOW_ORDER,
        motif_length_to_annotate=motif_length,
    )

    # test on vcf file (from somatic DV)
    annotated_somatic_dv_vcf = pjoin(tmpdir, "annotated_somatic_dv.vcf.gz")
    variant_annotation.VcfAnnotator.process_vcf(
        input_path=sample_vcf,
        output_path=annotated_somatic_dv_vcf,
        annotators=[ref_contetxt_variant_annotator],
    )
    __ref_context_annotator_assertion(
        annotated_vcf=annotated_somatic_dv_vcf,
        motif_length=motif_length,
        num_cycle_skips=3,
        num_non_cycle_skips=9,
        hmer_context_ref_sum=27,
        hmer_context_alt_sum=23,
    )


def test_ref_context_vcf_annotator_featuremap_with_bs(tmpdir, resources_dir):
    # test on FeatureMap file (from bsDNA)
    sample_featuremap = pjoin(resources_dir, "Pa_46.bsDNA.chr20_sample.vcf.gz")
    motif_length = 3
    ref_contetxt_variant_annotator = RefContextVcfAnnotator(
        ref_fasta=pjoin(resources_dir, "sample.fasta"),
        flow_order=DEFAULT_FLOW_ORDER,
        motif_length_to_annotate=motif_length,
    )
    annotated_balanced_featuremap = pjoin(tmpdir, "annotated_balanced_featuremap.vcf.gz")
    variant_annotation.VcfAnnotator.process_vcf(
        input_path=sample_featuremap,
        output_path=annotated_balanced_featuremap,
        annotators=[ref_contetxt_variant_annotator],
    )
    __ref_context_annotator_assertion(
        annotated_vcf=annotated_balanced_featuremap,
        motif_length=motif_length,
        num_cycle_skips=0,
        num_non_cycle_skips=31,
        hmer_context_ref_sum=64,
        hmer_context_alt_sum=41,
    )


def test_ref_context_vcf_annotator_multi_processing(tmpdir, resources_dir):
    # test on FeatureMap file, with multi-processing)
    sample_featuremap = pjoin(resources_dir, "Pa_46.bsDNA.chr20_sample.vcf.gz")
    motif_length = 3
    ref_contetxt_variant_annotator = RefContextVcfAnnotator(
        ref_fasta=pjoin(resources_dir, "sample.fasta"),
        flow_order=DEFAULT_FLOW_ORDER,
        motif_length_to_annotate=motif_length,
    )
    annotated_balanced_featuremap = pjoin(tmpdir, "annotated_balanced_featuremap.vcf.gz")
    variant_annotation.VcfAnnotator.process_vcf(
        input_path=sample_featuremap,
        output_path=annotated_balanced_featuremap,
        annotators=[ref_contetxt_variant_annotator],
        process_number=1,
    )
    __ref_context_annotator_assertion(
        annotated_vcf=annotated_balanced_featuremap,
        motif_length=motif_length,
        num_cycle_skips=0,
        num_non_cycle_skips=31,
        hmer_context_ref_sum=64,
        hmer_context_alt_sum=41,
    )


def __ref_context_annotator_assertion(
        annotated_vcf: str,
        motif_length: int,
        num_cycle_skips: int,
        num_non_cycle_skips: int,
        hmer_context_ref_sum: int,
        hmer_context_alt_sum: int,
):
    annotated_variants = pysam.VariantFile(annotated_vcf, "r")
    assert FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value in annotated_variants.header.info
    assert FeatureMapFields.HMER_CONTEXT_REF.value in annotated_variants.header.info
    assert FeatureMapFields.HMER_CONTEXT_ALT.value in annotated_variants.header.info
    assert FeatureMapFields.IS_CYCLE_SKIP.value in annotated_variants.header.info
    for i in range(motif_length):
        assert f"prev_{i + 1}" in annotated_variants.header.info
        assert f"next_{i + 1}" in annotated_variants.header.info

    cycle_skips = 0
    non_cycle_skips = 0
    hmer_context_ref_sum = 0
    hmer_context_alt_sum = 0
    for variant in annotated_variants:
        assert len(variant.info[FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value]) == motif_length + 1
        assert len(variant.info[FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value]) == motif_length + 1
        if variant.info[FeatureMapFields.IS_CYCLE_SKIP.value]:
            cycle_skips += 1
        else:
            non_cycle_skips += 1
        hmer_context_ref_sum += variant.info[FeatureMapFields.HMER_CONTEXT_REF.value]
        hmer_context_alt_sum += variant.info[FeatureMapFields.HMER_CONTEXT_ALT.value]

    assert cycle_skips == num_cycle_skips
    assert non_cycle_skips == num_non_cycle_skips
    assert hmer_context_ref_sum == hmer_context_ref_sum
    assert hmer_context_alt_sum == hmer_context_alt_sum



import logging
import os
import re
import subprocess
import tempfile
from collections import defaultdict
from os.path import basename, dirname
from os.path import join as pjoin

import numpy as np
import pysam

from ugbio_core.logger import logger
from ugbio_core.consts import AD, DP, GT, VAF
from ugbio_featuremap.featuremap_utils import FeatureMapFields, FeatureMapFilters
from ugbio_ppmseq.ppmSeq_consts import HistogramColumnNames

fields_to_collect_all_options = {
    "numeric_array_fields": [
        FeatureMapFields.X_SCORE.value,
        FeatureMapFields.X_EDIST.value,
        FeatureMapFields.X_LENGTH.value,
        FeatureMapFields.X_MAPQ.value,
        FeatureMapFields.X_INDEX.value,
        FeatureMapFields.X_FC1.value,
        FeatureMapFields.X_FC2.value,
        FeatureMapFields.MAX_SOFTCLIP_LENGTH.value,
        FeatureMapFields.X_FLAGS.value,
        HistogramColumnNames.STRAND_RATIO_START.value,
        HistogramColumnNames.STRAND_RATIO_END.value,
        "ML_QUAL",
    ],
    "string_list_fields": [
        FeatureMapFields.X_RN.value,
        FeatureMapFields.X_CIGAR.value,
        "rq",
        "tm",
        HistogramColumnNames.STRAND_RATIO_CATEGORY_START.value,
        HistogramColumnNames.STRAND_RATIO_CATEGORY_END.value,
        "st",
        "et",
    ],
    "boolean_fields": [
        FeatureMapFields.IS_FORWARD.value,
        FeatureMapFields.IS_DUPLICATE.value,
    ],
    "fields_to_write_once": [
        FeatureMapFields.READ_COUNT.value,
        FeatureMapFields.FILTERED_COUNT.value,
        FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value,
        FeatureMapFields.HMER_CONTEXT_REF.value,
        FeatureMapFields.HMER_CONTEXT_ALT.value,
        FeatureMapFields.PREV_1.value,
        FeatureMapFields.PREV_2.value,
        FeatureMapFields.PREV_3.value,
        FeatureMapFields.NEXT_1.value,
        FeatureMapFields.NEXT_2.value,
        FeatureMapFields.NEXT_3.value,
    ],
    "boolean_fields_to_write_once": [
        FeatureMapFields.IS_CYCLE_SKIP.value,
    ],
}


def write_a_pileup_record(
    record_dict: dict,
    rec_id: tuple,
    out_fh: pysam.VariantFile,
    header: pysam.VariantHeader,
    fields_to_collect: dict,
    min_qual: int,
    sample_name: str = "SAMPLE",
    qual_agg_func: callable = np.max,
):
    """
    Write a pileup record to a vcf file

    Inputs:
        record_dict (dict): A dictionary containing the record information
        rec_id (str): The record id
        out_fh (pysam.VariantFile): A pysam file handle of the output vcf file
        header (pysam.VariantHeader): A pysam VarianyHeader pbject of the output vcf file
        fields_to_collect (dict): A dictionary containing the fields to collect
        min_qual (int): The minimum quality threshold
        sample_name (str): The name of the sample (default: "SAMPLE")
        qual_agg_func (function): The function to aggregate the quality scores (default: np.max)
    Output:
        rec (pysam.VariantRecord): A pysam VariantRecord object

    """
    format_fields = [GT, AD, DP, VAF]
    rec = header.new_record()
    rec.chrom = rec_id[0]
    rec.pos = rec_id[1]
    rec.ref = rec_id[2]
    rec.alts = rec_id[3]
    rec.id = "."
    rec.qual = qual_agg_func(record_dict[FeatureMapFields.X_QUAL.value])
    # INFO fields to aggregate
    # exceptions: X_QUAL
    rec.info[FeatureMapFields.X_QUAL.value] = list(record_dict[FeatureMapFields.X_QUAL.value])
    for field in fields_to_collect["numeric_array_fields"]:
        if field in record_dict:
            rec.info[field] = list(record_dict[field])
    for field in fields_to_collect["string_list_fields"]:
        if field in record_dict:
            rec.info[field] = "|".join(record_dict[field])
    for field in fields_to_collect["boolean_fields"]:
        if field in record_dict:
            rec.info[field] = "|".join(["T" if f else "F" for f in record_dict[field]])
    for field in fields_to_collect["fields_to_write_once"]:
        if field in record_dict:
            rec.info[field] = record_dict[field][0]
            if not all(f == record_dict[field][0] for f in record_dict[field]):
                raise ValueError(
                    f"Field {field} has multiple values, but expected to have only a single value. rec: {record_dict}"
                )
    for field in fields_to_collect["boolean_fields_to_write_once"]:
        if field in record_dict:
            rec.info[field] = record_dict[field][0]
            if not all(f == record_dict[field][0] for f in record_dict[field]):
                raise ValueError(
                    f"Field {field} has multiple values, but expected to have only a single value. rec: {record_dict}"
                )

    # FORMAT fields to aggregate
    record_dict[DP] = rec.info[FeatureMapFields.FILTERED_COUNT.value]
    record_dict[VAF] = record_dict["rec_counter"] / record_dict[DP]
    record_dict[GT] = (0, 1)
    record_dict[AD] = [
        record_dict[DP] - record_dict["rec_counter"],
        record_dict["rec_counter"],
    ]
    for format_field in format_fields:
        rec.samples[sample_name][format_field] = record_dict[format_field]
    # FILTER field
    if rec.qual < min_qual:
        curr_filter = FeatureMapFilters.LOW_QUAL.value
    elif rec.qual >= min_qual and record_dict["rec_counter"] == 1:
        curr_filter = FeatureMapFilters.SINGLE_READ.value
    else:
        curr_filter = FeatureMapFilters.PASS.value
    rec.filter.add(curr_filter)
    # write to file
    out_fh.write(rec)
    return rec


def pileup_featuremap(  # pylint: disable=too-many-branches
    featuremap: str,
    output_vcf: str,
    genomic_interval: str = None,
    min_qual: int = 0,
    sample_name: str = "SAMPLE",
    qual_agg_func: callable = np.max,
    verbose: bool = True,
):
    """
    Pileup featuremap vcf to a regular vcf, save the aggregated quality and read count

    Inputs:
        featuremap (str): The input featuremap vcf file
        output_vcf (str): The output pileup vcf file
        genomic_interval (str): The genomic interval to pileup, format: chr:start-end
        min_qual (int): The minimum quality threshold
        sample_name (str): The name of the sample (default: "SAMPLE")
        qual_agg_func (callable): The function to aggregate the quality scores (default: np.max)
        verbose (bool): The verbosity level (default: True)
    Output:
        output_vcf (str): The output vcf file
    """
    if verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)
    # process genomic interval input
    if genomic_interval is None:
        chrom = None
        start = None
        end = None
    else:
        assert re.match(r"^\w+:\d+-\d+$", genomic_interval), "Input genomic_interval format should be 'chrom:start-end'"
        genomic_interval_list = genomic_interval.split(":")
        chrom = genomic_interval_list[0]
        start = int(genomic_interval_list[1].split("-")[0])
        end = int(genomic_interval_list[1].split("-")[1])

    # generate a new header
    # filter out the fields that are not present in the input featuremap vcf
    fields_to_collect = defaultdict(list)
    orig_header = pysam.VariantFile(featuremap).header
    header = pysam.VariantHeader()
    header.add_meta("fileformat", "VCFv4.2")
    header.filters.add(
        "SingleRead", None, None, "Aggregated quality above threshold, Only a single read agrees with alternative"
    )
    header.filters.add("LowQual", None, None, "Aggregated quality below threshold")

    for field in fields_to_collect_all_options["numeric_array_fields"]:
        if field in orig_header.info:
            header.info.add(field, ".", orig_header.info[field].type, orig_header.info[field].description)
            fields_to_collect["numeric_array_fields"] += [field]
        else:
            logger.debug(f"Field {field} not found in the input featuremap vcf")

    for field in fields_to_collect_all_options["string_list_fields"]:
        if field in orig_header.info:
            header.info.add(field, "1", orig_header.info[field].type, orig_header.info[field].description)
            fields_to_collect["string_list_fields"] += [field]
        else:
            logger.debug(f"Field {field} not found in the input featuremap vcf")

    for field in fields_to_collect_all_options["boolean_fields"]:
        if field in orig_header.info:
            header.info.add(field, "1", "String", orig_header.info[field].description)
            fields_to_collect["boolean_fields"] += [field]
        else:
            logger.debug(f"Field {field} not found in the input featuremap vcf")

    for field in fields_to_collect_all_options["fields_to_write_once"]:
        if field in orig_header.info:
            header.info.add(field, "1", orig_header.info[field].type, orig_header.info[field].description)
            fields_to_collect["fields_to_write_once"] += [field]
        else:
            logger.debug(f"Field {field} not found in the input featuremap vcf")

    for field in fields_to_collect_all_options["boolean_fields_to_write_once"]:
        if field in orig_header.info:
            header.info.add(field, "0", "Flag", orig_header.info[field].description)
            fields_to_collect["boolean_fields_to_write_once"] += [field]
        else:
            logger.debug(f"Field {field} not found in the input featuremap vcf")

    # exceptions: X_QUAL
    header.info.add(FeatureMapFields.X_QUAL.value, ".", "Float", "Quality of reads containing this location")
    header.formats.add(GT, 1, "String", "Genotype")
    header.formats.add(DP, 1, "Integer", "Read depth")
    header.formats.add(AD, "R", "Integer", "Allelic depths")
    header.formats.add(VAF, 1, "Float", "Allele frequency")
    # copy contigs from original header
    for contig in orig_header.contigs:
        header.contigs.add(contig)
    header.add_sample(sample_name)

    # open an output file
    out_fh = pysam.VariantFile(output_vcf, "w", header=header)

    # sort the vcf file prior to reading
    with tempfile.TemporaryDirectory(dir=dirname(output_vcf)) as temp_dir:
        sorted_featuremap = pjoin(temp_dir, basename(featuremap).replace(".vcf.gz", ".sorted.vcf.gz"))
        if genomic_interval is None:
            # sort all featuremap
            sort_cmd = f"bcftools sort {featuremap} -Oz -o {sorted_featuremap} && bcftools index -t {sorted_featuremap}"
        else:
            # sort only the genomic interval of interest
            sort_cmd = f"bcftools view {featuremap} {genomic_interval} |\
                  bcftools sort - -Oz -o {sorted_featuremap} && bcftools index -t {sorted_featuremap}"
        logger.debug(sort_cmd)
        subprocess.check_call(sort_cmd, shell=True)

        cons_dict = defaultdict(dict)
        with pysam.VariantFile(sorted_featuremap) as f:
            prev_key = tuple()
            for rec in f.fetch(chrom, start, end):
                rec_id = (rec.chrom, rec.pos, rec.ref, rec.alts[0])
                if "rec_counter" not in cons_dict[rec_id]:
                    if len(cons_dict.keys()) > 1:
                        # write to file
                        write_a_pileup_record(
                            cons_dict[prev_key],
                            prev_key,
                            out_fh,
                            header,
                            fields_to_collect,
                            min_qual,
                            sample_name,
                            qual_agg_func,
                        )
                        cons_dict.pop(prev_key)
                    # initialize rec_counter
                    cons_dict[rec_id]["rec_counter"] = 0
                    # exceptions: X_QUAL
                    cons_dict[rec_id][FeatureMapFields.X_QUAL.value] = np.array([])
                    for field in fields_to_collect["numeric_array_fields"]:
                        cons_dict[rec_id][field] = np.array([])
                    for field in fields_to_collect["string_list_fields"]:
                        cons_dict[rec_id][field] = []
                    for field in fields_to_collect["boolean_fields"]:
                        cons_dict[rec_id][field] = []
                    for field in fields_to_collect["fields_to_write_once"]:
                        cons_dict[rec_id][field] = []
                    for field in fields_to_collect["boolean_fields_to_write_once"]:
                        cons_dict[rec_id][field] = []

                # update the record
                cons_dict[rec_id]["rec_counter"] += 1
                # exceptions: X_QUAL
                cons_dict[rec_id][FeatureMapFields.X_QUAL.value] = np.append(
                    cons_dict[rec_id][FeatureMapFields.X_QUAL.value], rec.qual
                )
                for field in fields_to_collect["numeric_array_fields"]:
                    cons_dict[rec_id][field] = np.append(cons_dict[rec_id][field], rec.info.get(field, np.nan))
                for field in fields_to_collect["string_list_fields"]:
                    cons_dict[rec_id][field] += [rec.info.get(field, ".")]
                for field in fields_to_collect["boolean_fields"]:
                    cons_dict[rec_id][field] += [rec.info.get(field, False)]
                for field in fields_to_collect["fields_to_write_once"]:
                    cons_dict[rec_id][field] += [rec.info.get(field, None)]
                for field in fields_to_collect["boolean_fields_to_write_once"]:
                    cons_dict[rec_id][field] += [rec.info.get(field, False)]
                prev_key = rec_id

            # write last record
            if len(cons_dict.keys()) > 0:
                write_a_pileup_record(
                    cons_dict[prev_key],
                    prev_key,
                    out_fh,
                    header,
                    fields_to_collect,
                    min_qual,
                    sample_name,
                    qual_agg_func,
                )

        out_fh.close()
        pysam.tabix_index(output_vcf, preset="vcf", min_shift=0, force=True)
    return output_vcf


def pileup_featuremap_on_an_interval_list(
    featuremap: str,
    output_vcf: str,
    interval_list: str,
    min_qual: int = 0,
    sample_name: str = "SAMPLE",
    qual_agg_func: callable = np.max,
    verbose: bool = True,
):
    """
    Apply pileup featuremap on an interval list

    Inputs:
        featuremap (str): The input featuremap vcf file
        output_vcf (str): The output pileup vcf file
        interval_list (str): The interval list file
        min_qual (int): The minimum quality threshold
        sample_name (str): The name of the sample (default: "SAMPLE")
        qual_agg_func (callable): The function to aggregate the quality scores (default: np.max)
        verbose (bool): The verbosity level (default: True)
    Output:
        output_vcf (str): The output vcf file
    """
    with tempfile.TemporaryDirectory(dir=dirname(output_vcf)) as temp_dir:
        with open(interval_list, "r", encoding="utf-8") as f:
            for line in f:
                # ignore header lines
                if line.startswith("@"):
                    continue
                # read genomic ineterval
                genomic_interval = line.strip()
                genomic_interval_list = genomic_interval.split("\t")
                chrom = genomic_interval_list[0]
                start = genomic_interval_list[1]
                end = genomic_interval_list[2]
                genomic_interval = chrom + ":" + str(start) + "-" + str(end)
                # run pileup_featuremap on the interval
                curr_output_vcf = pjoin(
                    temp_dir,
                    basename(output_vcf).replace(".vcf.gz", "")
                    + "."
                    + chrom
                    + "_"
                    + str(start)
                    + "_"
                    + str(end)
                    + ".int_list.vcf.gz",
                )
                pileup_featuremap(
                    featuremap, curr_output_vcf, genomic_interval, min_qual, sample_name, qual_agg_func, verbose
                )
        # merge the output vcfs
        vcfs = [pjoin(temp_dir, f) for f in os.listdir(temp_dir) if f.endswith(".vcf.gz")]
        vcf_str = " ".join(vcfs)
        cmd = f"bcftools concat {vcf_str} -a | bcftools sort - -Oz -o {output_vcf} && bcftools index -t {output_vcf}"
        logger.debug(cmd)
        subprocess.check_call(cmd, shell=True)
    return output_vcf



import itertools
import os
from enum import Enum
from os.path import join as pjoin, dirname, splitext, basename

import numpy as np
import pandas as pd
import pyfaidx
import pysam
from simppl.simple_pipeline import SimplePipeline

from ugbio_core.logger import logger
from ugbio_core.consts import ALT, CHROM, DEFAULT_FLOW_ORDER, FILTER, IS_CYCLE_SKIP, POS, QUAL, REF
from ugbio_core.dna_sequence_utils import get_max_softclip_len
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.sorter_utils import read_effective_coverage_from_sorter_json
from ugbio_core.variant_annotation import VcfAnnotator, get_cycle_skip_dataframe, get_motif_around_snv
from ugbio_core.consts import FileExtension
from ugbio_ppmseq.ppmSeq_utils import ppmSeqStrandVcfAnnotator


class FeatureMapFields(Enum):
    CHROM = CHROM
    POS = POS
    REF = REF
    ALT = ALT
    QUAL = QUAL
    FILTER = FILTER
    READ_COUNT = "X_READ_COUNT"
    FILTERED_COUNT = "X_FILTERED_COUNT"
    X_SCORE = "X_SCORE"
    X_EDIST = "X_EDIST"
    X_LENGTH = "X_LENGTH"
    X_MAPQ = "X_MAPQ"
    X_INDEX = "X_INDEX"
    X_FC1 = "X_FC1"
    X_FC2 = "X_FC2"
    X_QUAL = "X_QUAL"
    X_RN = "X_RN"
    TRINUC_CONTEXT_WITH_ALT = "trinuc_context_with_alt"
    HMER_CONTEXT_REF = "hmer_context_ref"
    HMER_CONTEXT_ALT = "hmer_context_alt"
    IS_CYCLE_SKIP = IS_CYCLE_SKIP
    IS_FORWARD = "is_forward"
    IS_DUPLICATE = "is_duplicate"
    MAX_SOFTCLIP_LENGTH = "max_softclip_length"
    X_FLAGS = "X_FLAGS"
    X_CIGAR = "X_CIGAR"
    PREV_1 = "prev_1"
    PREV_2 = "prev_2"
    PREV_3 = "prev_3"
    NEXT_1 = "next_1"
    NEXT_2 = "next_2"
    NEXT_3 = "next_3"


class FeatureMapFilters(Enum):
    LOW_QUAL = "LowQual"
    SINGLE_READ = "SingleRead"
    PASS = "PASS"
    PRE_FILTERED = "PreFiltered"


def get_hmer_of_central_base(sequence: str) -> int:
    """
    Get the length of the homopolymer spanning the central base in the given sequence.
    Examples:
    ACA -> 1
    ACC -> 2
    ACCCT -> 3
    AAAAAT -> 5

    Parameters
    ----------
    sequence : str
        The sequence to check. Must be of odd length, so the central base is well defined.

    Returns
    -------
    int
        The length of the homopolymer run of the central base.
    """
    assert isinstance(sequence, str)
    assert len(sequence) % 2 == 1
    assert len(sequence) >= 1
    hmer_lengths = np.array([sum(1 for _ in x[1]) for x in itertools.groupby(sequence)])
    central_hmer = hmer_lengths[np.argmax(np.cumsum(hmer_lengths) > len(sequence) // 2)]
    return int(central_hmer)


def is_biallelic_snv(record: pysam.VariantRecord) -> bool:
    """
    Check if a given VCF record is a biallelic SNV.

    Parameters
    ----------
    record : pysam.VariantRecord
        The VCF record to check.

    Returns
    -------
    bool
        True if the record is a biallelic SNV, False otherwise.
    """
    assert isinstance(record, pysam.VariantRecord)
    return len(record.ref) == 1 and len(record.alts) == 1 and len(record.alts[0]) == 1


class FeaturemapAnnotator(VcfAnnotator):
    """
    Annotate vcf with featuremap-specific fields derived from X_FLAGS and X_CIGAR:
    - is_forward: is the read forward mapped
    - is_duplicate: is the read a duplicate
    - max_softclip_length: maximum softclip length
    """

    def __init__(self):
        pass

    def edit_vcf_header(self, header: pysam.VariantHeader) -> pysam.VariantHeader:
        """
        Edit the VCF header to include new fields

        Parameters
        ----------
        header : pysam.VariantHeader
            VCF header

        Returns
        -------
        pysam.VariantHeader
            Modified VCF header

        """
        header.add_line("##ugbio_mrd.mrd_utils.py_FeaturemapAnnotator=.")
        header.add_line(
            f"##INFO=<ID={FeatureMapFields.IS_FORWARD.value},"
            'Number=0,Type=Flag,Description="is the read forward mapped">'
        )
        header.add_line(
            f"##INFO=<ID={FeatureMapFields.IS_DUPLICATE.value},"
            'Number=0,Type=Flag,Description="is the read a duplicate">'
        )
        header.add_line(
            f"##INFO=<ID={FeatureMapFields.MAX_SOFTCLIP_LENGTH.value},"
            'Number=1,Type=Integer,Description="maximum softclip length between start and end of read">'
        )

        return header

    def process_records(self, records: list[pysam.VariantRecord]) -> list[pysam.VariantRecord]:
        """

        Parameters
        ----------
        records : list[pysam.VariantRecord]
            list of VCF records

        Returns
        -------
        list[pysam.VariantRecord]
            list of updated VCF records
        """
        records_out = [None] * len(records)
        for j, record in enumerate(records):
            if FeatureMapFields.X_FLAGS.value in record.info:
                flags = record.info[FeatureMapFields.X_FLAGS.value]
                record.info[FeatureMapFields.IS_FORWARD.value] = (flags & 16) == 0
                record.info[FeatureMapFields.IS_DUPLICATE.value] = (flags & 1024) != 0
            if FeatureMapFields.X_CIGAR.value in record.info:
                record.info[FeatureMapFields.MAX_SOFTCLIP_LENGTH.value] = get_max_softclip_len(
                    record.info[FeatureMapFields.X_CIGAR.value]
                )
            records_out[j] = record

        return records_out


# pylint: disable=too-many-instance-attributes
class RefContextVcfAnnotator(VcfAnnotator):
    def __init__(
        self,
        ref_fasta: str,
        flow_order: str = DEFAULT_FLOW_ORDER,
        motif_length_to_annotate: int = 3,
        max_hmer_length: int = 20,
    ):
        """
        Annotator to add reference context to VCF records, only modifies biallelic SNVs.
        The following are added to the INFO field:
        - trinuc_context_with_alt: reference trinucleotide context
        - prev_N: base i in the reference before the variant, i in range 1 to N=length motif_length_to_annotate
        - next_N: base i in the reference after the variant, i in range 1 to N=length motif_length_to_annotate
        - hmer_context_ref: reference homopolymer context, up to length max_hmer_length
        - hmer_context_alt: homopolymer context in the ref allele (assuming the variant considered only),
            up to length max_hmer_length
        - is_cycle_skip: True if the SNV is a cycle skip

        Parameters
        ----------
        ref_fasta : str
            Path to the reference FASTA file.
        flow_order : str
            Flow order of the flow cell.
        motif_length_to_annotate : int, optional
            The length of the motif to annotate context up to (prev_N / next_N). Defaults to 3.
        max_hmer_length : int, optional
            The maximum length of the homopolymer to annotate context up to. Defaults to 20.
        """
        # check inputs
        assert len(flow_order) == 4, f"Flow order must be of length 4, got {flow_order}"
        assert os.path.isfile(ref_fasta), f"Reference FASTA file not found: {ref_fasta}"

        # save inputs
        self.ref_fasta = ref_fasta
        self.motif_length_to_annotate = motif_length_to_annotate
        self.max_hmer_length = max_hmer_length
        self.flow_order = flow_order

        # init accesory objects
        self.cycle_skip_dataframe = get_cycle_skip_dataframe(flow_order)

        # info field names
        self.TRINUC_CONTEXT_WITH_ALT = FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value
        self.HMER_CONTEXT_REF = FeatureMapFields.HMER_CONTEXT_REF.value
        self.HMER_CONTEXT_ALT = FeatureMapFields.HMER_CONTEXT_ALT.value
        self.CYCLE_SKIP_FLAG = FeatureMapFields.IS_CYCLE_SKIP.value
        self.SINGLE_REF_BASES = []
        for i in range(self.motif_length_to_annotate):
            self.SINGLE_REF_BASES.append(f"prev_{i + 1}")
            self.SINGLE_REF_BASES.append(f"next_{i + 1}")

        self.info_fields_to_add = [
            self.TRINUC_CONTEXT_WITH_ALT,
            self.HMER_CONTEXT_REF,
            self.HMER_CONTEXT_ALT,
        ] + self.SINGLE_REF_BASES
        # self.CYCLE_SKIP_FLAG not included because it's a flag, only there if True

    def edit_vcf_header(self, header: pysam.VariantHeader) -> pysam.VariantHeader:
        """
        Edit the VCF header to include new fields

        Parameters
        ----------
        header : pysam.VariantHeader
            VCF header

        Returns
        -------
        pysam.VariantHeader
            Modified VCF header

        """
        header.add_line(
            "##ugbio_core.variant_annotation._RefContextVcfAnnotator="
            f"ref:{os.path.basename(self.ref_fasta)}"
            f"_motif_length_to_annotate:{self.motif_length_to_annotate}"
            f"_max_hmer_length:{self.max_hmer_length}"
        )

        header.add_meta(
            "INFO",
            items=[
                ("ID", self.TRINUC_CONTEXT_WITH_ALT),
                ("Number", 1),
                ("Type", "String"),
                ("Description", "reference trinucleotide context and alt base"),
            ],
        )

        for i in range(self.motif_length_to_annotate):
            header.add_meta(
                "INFO",
                items=[
                    ("ID", f"prev_{i + 1}"),
                    ("Number", 1),
                    ("Type", "String"),
                    ("Description", f"{i + 1} bases in the reference before variant"),
                ],
            )
            header.add_meta(
                "INFO",
                items=[
                    ("ID", f"next_{i + 1}"),
                    ("Number", 1),
                    ("Type", "String"),
                    ("Description", f"{i + 1} bases in the reference after variant"),
                ],
            )

        header.add_meta(
            "INFO",
            items=[
                ("ID", self.HMER_CONTEXT_REF),
                ("Number", 1),
                ("Type", "Integer"),
                ("Description", f"reference homopolymer context, up to length {self.max_hmer_length}"),
            ],
        )

        header.add_meta(
            "INFO",
            items=[
                ("ID", self.HMER_CONTEXT_ALT),
                ("Number", 1),
                ("Type", "Integer"),
                (
                    "Description",
                    f"homopolymer context in the ref allele (assuming the variant considered only), "
                    f"up to length {self.max_hmer_length}",
                ),
            ],
        )

        header.add_meta(
            "INFO",
            items=[
                ("ID", self.CYCLE_SKIP_FLAG),
                ("Number", 0),
                ("Type", "Flag"),
                ("Description", "True if the SNV is a cycle skip"),
            ],
        )

        return header

    def process_records(self, records: list[pysam.VariantRecord]) -> list[pysam.VariantRecord]:
        """
        Parameters
        ----------
        records : list[pysam.VariantRecord]
            list of VCF records

        Returns
        -------
        list[pysam.VariantRecord]
            list of updated VCF records
        """
        records_out = [None] * len(records)
        faidx_ref = pyfaidx.Fasta(self.ref_fasta, build_index=False, rebuild=False)
        for j, record in enumerate(records):
            if is_biallelic_snv(record):
                # get motif
                ref_around_snv = get_motif_around_snv(record, self.max_hmer_length, faidx_ref)
                central_base_ind = len(ref_around_snv) // 2
                trinuc_ref = ref_around_snv[central_base_ind - 1 : central_base_ind + 2]
                # create sequence of alt allele
                alt_around_snv_list = list(ref_around_snv)
                alt_around_snv_list[central_base_ind] = record.alts[0]
                alt_around_snv = "".join(alt_around_snv_list)
                trinuc_alt = alt_around_snv[central_base_ind - 1 : central_base_ind + 2]
                # assign to record
                record.info[self.HMER_CONTEXT_REF] = get_hmer_of_central_base(ref_around_snv)
                record.info[self.HMER_CONTEXT_ALT] = get_hmer_of_central_base(alt_around_snv)
                record.info[self.TRINUC_CONTEXT_WITH_ALT] = trinuc_ref + record.alts[0]

                for index in range(1, self.motif_length_to_annotate + 1):
                    field_name = f"next_{index}"
                    record.info[field_name] = ref_around_snv[central_base_ind + index]
                    field_name = f"prev_{index}"
                    record.info[field_name] = ref_around_snv[central_base_ind - index]

                is_cycle_skip = self.cycle_skip_dataframe.at[(trinuc_ref, trinuc_alt), IS_CYCLE_SKIP]
                record.info[self.CYCLE_SKIP_FLAG] = is_cycle_skip

                # make sure all the info fields are present
                for info_field in self.info_fields_to_add:
                    assert (
                        info_field in record.info
                    ), f"INFO field {info_field} was supposed to be added to VCF record but was not found"
            records_out[j] = record

        return records_out


def create_hom_snv_featuremap(
    featuremap: str,
    sorter_stats_json: str = None,
    hom_snv_featuremap: str = None,
    sp: SimplePipeline = None,
    requested_min_coverage: int = 20,
    min_af: float = 0.7,
):
    """Create a HOM SNV featuremap from a featuremap

    Parameters
    ----------
    featuremap : str
        Input featuremap.
    sorter_stats_json : str
        Path to Sorter statistics JSON file, used to extract the median coverage. If None (default), minimum coverage
        will be set to requested_min_coverage even if the median coverage is lower, might yield an empty output.
    hom_snv_featuremap : str, optional
        Output featuremap with HOM SNVs reads to be used as True Positives. If None (default),
        the hom_snv_featuremap will be the same as the input featuremap with a ".hom_snv.vcf.gz" suffix.
    sp : SimplePipeline, optional
        SimplePipeline object to use for running commands. If None (default), commands will be run using subprocess.
    requested_min_coverage : int, optional
        Minimum coverage requested for locus to be propagated to the output. If the median coverage is lower than this
        value, the median coverage will be used as the minimum coverage instead.
        Default 20
    min_af : float, optional
        Minimum allele fraction in the featuremap to be considered a HOM SNV
        Default 0.7
        The default is chosen as 0.7 and not higher because some SNVs are pre-filtered from the FeatureMap due to
        MAPQ<60 or due to adjacent hmers.
    """

    # check inputs
    assert os.path.isfile(featuremap), f"featuremap {featuremap} does not exist"
    if sorter_stats_json:
        assert os.path.isfile(sorter_stats_json), f"sorter_stats_json {sorter_stats_json} does not exist"
    if hom_snv_featuremap is None:
        if featuremap.endswith(".vcf.gz"):
            hom_snv_featuremap = featuremap[: -len(".vcf.gz")]
        hom_snv_featuremap = featuremap + ".hom_snv.vcf.gz"
    hom_snv_bed = hom_snv_featuremap.replace(".vcf.gz", ".bed.gz")
    logger.info(f"Writting HOM SNV featuremap to {hom_snv_featuremap}")

    # get minimum coverage
    if sorter_stats_json:
        (
            _,
            _,
            _,
            min_coverage,
            _,
        ) = read_effective_coverage_from_sorter_json(sorter_stats_json, min_coverage_for_fp=requested_min_coverage)
    else:
        min_coverage = requested_min_coverage
    logger.info(
        f"Using a minimum coverage of {min_coverage} for HOM SNV featuremap (requested {requested_min_coverage})"
    )

    # Create commands to filter the featuremap for homozygous SNVs.
    cmd_get_hom_snv_loci_bed_file = (
        # Use bcftools to query specific fields in the vcf file. This includes the chromosome (CHROM),
        # the 0-based start position (POS0), the 1-based start position (POS), and the number of reads
        # in the locus (X_READ_COUNT) for the specified feature map.
        f"bcftools query -f '%CHROM\t%POS0\t%POS\t%INFO/{FeatureMapFields.READ_COUNT.value}\n' {featuremap} |"
        # Pipe the output to bedtools groupby command.
        # Here, -c 3 means we are specifying the third column as the key to groupby.
        # The '-full' option includes all columns from the input in the output.
        # The '-o count' option is specifying to count the number of lines for each group.
        f"bedtools groupby -c 3 -full -o count | "
        # Pipe the result to an awk command, which filters the result based on minimum coverage and allele frequency.
        # The '$4>=~{min_coverage}' part checks if the fourth column (which should be read count) is greater than or
        # equal to the minimum coverage. The '$5/$4>=~{min_af}' part checks if the allele frequency (calculated as
        # column 5 divided by column 4) is greater than or equal to the minimum allele frequency.
        f"awk '($4>={min_coverage})&&($5/$4>={min_af})' | "
        # The final output is then compressed and saved to the specified location in .bed.gz format.
        f"gzip > {hom_snv_bed}"
    )
    cmd_intersect_bed_file_with_original_featuremap = (
        f"bedtools intersect -a {featuremap} -b {hom_snv_bed} -u -header | bcftools view - -Oz -o {hom_snv_featuremap}"
    )
    cmd_index_hom_snv_featuremap = f"bcftools index -ft {hom_snv_featuremap}"

    # Run the commands
    try:
        for command in (
            cmd_get_hom_snv_loci_bed_file,
            cmd_intersect_bed_file_with_original_featuremap,
            cmd_index_hom_snv_featuremap,
        ):
            print_and_execute(command, simple_pipeline=sp, module_name=__name__)

    finally:
        # remove temp file
        if os.path.isfile(hom_snv_bed):
            os.remove(hom_snv_bed)


def filter_featuremap_with_bcftools_view(
    input_featuremap_vcf: str,
    intersect_featuremap_vcf: str,
    min_coverage: int = None,
    max_coverage: int = None,
    regions_file: str = None,
    bcftools_include_filter: str = None,
    sp: SimplePipeline = None,
) -> str:
    """
    Create a bcftools view command to filter a featuremap vcf

    Parameters
    ----------
    input_featuremap_vcf : str
        Path to input featuremap vcf
    intersect_featuremap_vcf : str
        Path to output intersected featuremap vcf
    min_coverage : int, optional
        Minimum coverage to include, by default None
    max_coverage : int, optional
        Maximum coverage to include, by default None
    regions_file : str, optional
        Path to regions file, by default None
    bcftools_include_filter: str, optional
        bcftools include filter to apply as part of a "bcftools view <vcf> -i 'pre_filter_bcftools_include'"
        before sampling, by default None
    sp : SimplePipeline, optional
        SimplePipeline object to use for printing and running commands, by default None
    """
    bcftools_view_command = f"bcftools view {input_featuremap_vcf} -O z -o {intersect_featuremap_vcf} "
    include_filters = [bcftools_include_filter.replace("'", '"')] if bcftools_include_filter else []
    # filter out variants with coverage outside of min and max coverage
    if min_coverage is not None:
        include_filters.append(f"(INFO/{FeatureMapFields.READ_COUNT.value} >= {min_coverage})")
    if max_coverage is not None:
        include_filters.append(f"(INFO/{FeatureMapFields.READ_COUNT.value} <= {max_coverage})")

    bcftools_include_string = " && ".join(include_filters)
    if bcftools_include_string:
        bcftools_view_command += f" -i '{bcftools_include_string}' "
    if regions_file:
        bcftools_view_command += f" -T {regions_file} "
    bcftools_index_command = f"bcftools index -t {intersect_featuremap_vcf}"

    print_and_execute(bcftools_view_command, simple_pipeline=sp, module_name=__name__, shell=True)
    print_and_execute(
        bcftools_index_command,
        simple_pipeline=sp,
        module_name=__name__,
    )
    assert os.path.isfile(intersect_featuremap_vcf), f"failed to create {intersect_featuremap_vcf}"


def featuremap_to_dataframe(
    featuremap_vcf: str,
    output_file: str = None,
    input_info_fields: list[str] = "all",
    input_format_fields: list[str] = None,
    sample_name: str = None,
    sample_index: int = None,
    default_int_fillna_value: int = 0,
    default_string_fillna_value: int = np.nan,
):
    """
    Converts featuremap in vcf format to dataframe

    Parameters
    ----------
    featuremap_vcf : str
        Path to featuremap vcf file
    output_file : str, optional
        Path to output file, by default None
        If None, saved to the same basename as featuremap_vcf with .parquet extension
    input_info_fields : list[str], optional
        List of input info fields to include in dataframe, by default 'all''
        If 'all' then all the info fields are read to columns
        If None then no info fields are read to columns
    input_format_fields : list[str], optional
        List of format info fields to include in dataframe, by default None
        If 'all' then all the info fields are read to columns
        If None then no info fields are read to columns
    sample_name : str, optional
        Name of sample to read formats columns from, default None
        If sample_name and sample_index are both None then no formats columns are read
    sample_index : int, optional
        Index of sample to read formats columns from, default None
        If sample_name and sample_index are both None and format_info_fields is not empty
        then sample_index defaults to 0
    default_int_fillna_value : int, optional
        Value to fill na values with for Integer fields, by default 0
    default_string_fillna_value : str, optional
        Value to fill na values with for String fields, by default np.nan

    Returns
    -------
    pd.DataFrame
        Dataframe of featuremap

    Raises
    ------
    ValueError
        If sample_name and sample_index are both not None
        If sample_name is not in vcf file
        If sample_index is out of range
        If input_info_fields contains a field not in vcf header
        If format_info_fields contains a field not in vcf header

    """
    # check inputs
    if sample_name is not None and sample_index is not None:
        raise ValueError("Cannot specify both sample_name and sample_index")
    if sample_name is not None and sample_name not in pysam.VariantFile(featuremap_vcf).header.samples:
        raise ValueError(f"Sample {sample_name} not in vcf file")
    if sample_index is not None and sample_index >= len(pysam.VariantFile(featuremap_vcf).header.samples):
        raise ValueError(f"Sample index {sample_index} out of range")
    if input_info_fields is not None and input_info_fields != "all" and len(input_info_fields) > 0:
        for info_field in input_info_fields:
            if info_field not in pysam.VariantFile(featuremap_vcf).header.info:
                raise ValueError(f"Input info field {info_field} not in vcf header")
    if input_format_fields is not None and input_format_fields != "all" and len(input_format_fields) > 0:
        for info_field in input_format_fields:
            if info_field not in pysam.VariantFile(featuremap_vcf).header.formats:
                raise ValueError(f"Format info field {info_field} not in vcf header")
        if sample_index is None and sample_name is None:
            sample_index = 0
        if sample_index is None and sample_name is not None:
            sample_index = list(pysam.VariantFile(featuremap_vcf).header.samples).index(sample_name)

    # define type conversion dictionary, String is converted to object to support np.nan
    type_conversion_dict = {
        "String": object,
        "Integer": int,
        "Float": float,
        "Flag": bool,
    }

    # auxiliary function to parse info and formats fields to read
    def get_metadata_dict(values: list[str], input_fields: list[str] = None):
        if input_fields is None:
            return {}, []
        metadata_dict_full = {value.name: (value.number, value.type) for value in values}
        input_fields_to_use = list(metadata_dict_full.keys()) if input_fields.lower() == "all" else input_fields
        for key in input_fields_to_use:
            if key not in metadata_dict_full:
                raise ValueError(f"Input field {key} missing from vcf header")
        metadata_dict = {key: metadata_dict_full[key] for key in input_fields_to_use}
        return metadata_dict, input_fields_to_use

    # read vcf file to dataframe
    with pysam.VariantFile(featuremap_vcf) as vcf_handle:
        # parse header and get metadata
        header = vcf_handle.header
        info_metadata_dict, info_input_fields_to_use = get_metadata_dict(
            values=header.info.values(), input_fields=input_info_fields
        )
        format_metadata_dict, format_input_fields_to_use = get_metadata_dict(
            values=header.formats.values(), input_fields=input_format_fields
        )
        # TODO remove this patch once the issue is fixed in FeatureMap
        # This section is a patch for specific integer tags, since there is an issue with FeautreMap
        # not propagating the types of copied SAM tags properly, so that they all default to String.
        # There is a pending request to fix, in the mean time this is a workaround.
        harcoded_info_fields = {x: "Integer" for x in ("a3", "ae", "as", "s2", "s3", "te", "ts")}
        harcoded_info_fields["rq"] = "Float"
        for field, datatype in harcoded_info_fields.items():
            if field in info_metadata_dict:
                info_metadata_dict[field] = (info_metadata_dict[field][0], datatype)
        # read vcf file to dataframe
        df = pd.DataFrame(
            vcf_row_generator(
                vcf_handle=vcf_handle,
                info_metadata_dict=info_metadata_dict,
                formats_metadata_dict=format_metadata_dict,
                sample_index=sample_index,
            ),
            columns=[CHROM, POS, REF, ALT, QUAL, FILTER] + info_input_fields_to_use + format_input_fields_to_use,
        )
    # fillna in int columns before converting types, because int columns can't contain NaNs
    df = df.fillna(
        {
            info_key: default_int_fillna_value
            for info_key in info_metadata_dict.keys()
            if type_conversion_dict[info_metadata_dict[info_key][1]] == int
        }
    )
    df.fillna(
        {
            info_key: default_string_fillna_value
            for info_key in info_metadata_dict.keys()
            if type_conversion_dict[info_metadata_dict[info_key][1]] == object
        }
    )
    # convert types
    df = df.astype(
        {
            **{"qual": float},
            **{
                info_key: type_conversion_dict[info_metadata_dict[info_key][1]]
                for info_key in info_metadata_dict.keys()
                if info_metadata_dict[info_key][0]
            },
        }
    )
    # Determine output file name
    if output_file is None:
        if featuremap_vcf.endswith(FileExtension.VCF_GZ.value):
            output_file = pjoin(
                dirname(featuremap_vcf),
                splitext(splitext(basename(featuremap_vcf))[0])[0] + FileExtension.PARQUET.value,
            )
        else:
            output_file = pjoin(
                dirname(featuremap_vcf),
                splitext(basename(featuremap_vcf))[0] + FileExtension.PARQUET.value,
            )
    elif not output_file.endswith(FileExtension.PARQUET.value):
        output_file = f"{output_file}{FileExtension.PARQUET.value}"

    # Save and return
    df.to_parquet(output_file)
    return df

def annotate_featuremap(
    input_featuremap: str,
    output_featuremap: str,
    ref_fasta: str,
    motif_length_to_annotate: int,
    max_hmer_length: int,
    flow_order: str = DEFAULT_FLOW_ORDER,
    ppmSeq_adapter_version: str = None,
    process_number: int = 0,
):
    """
    Annotate featuremap with ref context, hmer length and ppmSeq features

    Parameters
    ----------
    input_featuremap: str
        Input featuremap file
    output_featuremap: str
        Output featuremap file
    ref_fasta: str
        Reference fasta file
    motif_length_to_annotate: int
        Motif length to annotate
    max_hmer_length: int
        Max hmer length
    flow_order: str, optional
        Flow order, default TGCA
    ppmSeq_adapter_version: str, optional
        ppmSeq adapter version, if None no ppmSeq annotation is performed
    """
    featuremap_annotator = FeaturemapAnnotator()
    ref_context_annotator = RefContextVcfAnnotator(
        ref_fasta=ref_fasta,
        flow_order=flow_order,
        motif_length_to_annotate=motif_length_to_annotate,
        max_hmer_length=max_hmer_length,
    )
    annotators = [featuremap_annotator, ref_context_annotator]
    if ppmSeq_adapter_version:
        ppmSeq_annotator = ppmSeqStrandVcfAnnotator(adapter_version=ppmSeq_adapter_version)
        annotators.append(ppmSeq_annotator)
    VcfAnnotator.process_vcf(
        annotators=annotators,
        input_path=input_featuremap,
        output_path=output_featuremap,
        process_number=process_number,
    )


def vcf_row_generator(
    vcf_handle: pysam.VariantFile,
    info_metadata_dict: dict = None,
    formats_metadata_dict: dict = None,
    sample_index: int = None,
):
    """
    Return a generator object that reads entries from an open vcf file handle and yields a list of values

    Parameters
    ----------
    vcf_handle: pysam.VariantFile
        An open vcf file handle
    info_metadata_dict: dict
        A dictionary of info columns to read from the vcf file. The keys are the info column names and the values
        default None, in which case no INFO fields are read
    formats_metadata_dict: dict
        A dictionary of format columns to read from the vcf file. The keys are the info column names and the values
        default None, in which case no FORMAT fields are read
    sample_index: int
        The index of the sample to read from the vcf file.

    """
    for record in vcf_handle.fetch():
        # read basic fields
        entry = [
            record.chrom,
            record.pos,
            record.ref,
            record.alts[0] if len(record.alts) >= 1 else None,
            record.qual,
            ";".join(record.filter.keys()),
        ]
        # read info fields
        if info_metadata_dict is not None and len(info_metadata_dict) > 0:
            entry += [
                record.info.get(info_key)[0]
                if info_value[0] == "A"
                and record.info.get(info_key, None) is not None
                and len(record.info.get(info_key)) >= 1
                else record.info.get(info_key, np.nan)
                for info_key, info_value in info_metadata_dict.items()
            ]
        # read format fields
        if formats_metadata_dict is not None and len(formats_metadata_dict) > 0:
            entry += [
                record.samples[sample_index].get(formats_key, np.nan) for formats_key in formats_metadata_dict.keys()
            ]
        yield entry
